#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "spi.h"

static const char* device  = "/dev/spidev0.0";

spi_settings_t spi_settings_desired = {
    .mode          = SPI_MODE_3,
    .is_lsb_first  = 0,
    .bits_per_word = 8,
    .max_speed_hz  = 25000
};

int16_t read_mcp3301_single(int fd) {
    uint8_t raw_data[2];
    int16_t data = 0;
    char sign = 0;

    if (2 != spi_read_two_bytes(fd, raw_data)) {
        printf("read_mcp3301_single: failed to read value\n");
        return 0x8001; // Should be an invalid output for the sensor
    }

    data |= ((raw_data[0] & 0x0F) << 8);
    data |= raw_data[1];
    sign = (raw_data[0] & 0x10) ? 1 : 0;
    if (sign) {
        data -= 4096;
    }

    return data;
}

double average16(int16_t *vals, int cnt) {
    double val = 0;
    for(int i = 0; i < cnt; i++) {
        val += vals[i];
    }
    return val / cnt;
}

double stdev16(int16_t *vals, int cnt) {
    double stdev = 0;
    double avg = average16(vals, cnt);
    for(int i = 0; i < cnt; i++) {
        stdev += pow(vals[i] - avg, 2);
    }
    return sqrt(stdev / cnt);
}

typedef struct mcp3301_measurement {
    int16_t int_val;
    double  timestamp;
} mcp3301_measurement_t;

mcp3301_measurement_t read_mcp3301_measurement(int fd, clock_t time_init) {
    mcp3301_measurement_t mt = {0, 0.0};
    mt.int_val = read_mcp3301_single(fd);
    mt.timestamp = ((double)(clock() - time_init)) / CLOCKS_PER_SEC;
    return mt;
}

typedef struct filter_buffer {
    int    *data;
    size_t  data_len;
    int     location;
} filter_buffer_t;

filter_buffer_t *fb_new(size_t len) {
    filter_buffer_t *fb = (filter_buffer_t *) malloc(sizeof(filter_buffer_t));
    fb->data = (int*) malloc(sizeof(int) * len);
    memset(fb->data, 0, sizeof(int) * len);
    fb->data_len = len;
    fb->location = 0;
    return fb;
}

void fb_del(filter_buffer_t *fb) {
    free(fb->data);
    free(fb);
}

void fb_incr_loc(filter_buffer_t *fb) {
    fb->location = (fb->location + 1) % fb->data_len;
}

void fb_push(filter_buffer_t *fb, int val) {
    fb->data[fb->location] = val;
    fb_incr_loc(fb);
}

double filter_avg(filter_buffer_t *fb) {
    int max = fb->data[0];
    int min = fb->data[0];
    int sum = max;
    int data;
    for(int i = 1; i < fb->data_len; i++) {
        data = fb->data[i];
        if (data > max) {
            max = data;
        }
        if (data < min) {
            min = data;
        }
        sum += data;
    }
    sum = sum - max - min;
    return ((double) sum) / (fb->data_len - 2);
}

int main(int argc, char** argv) {
    clock_t t_init = clock();
    int spi_fd = 0;
    if (0 >= (spi_fd = spi_init(device, &spi_settings_desired))) {
        printf("main: could not initialize SPI bus\n");
        goto fail;
    }

    filter_buffer_t *fb = fb_new(16);

    int loops = 0;
    double t = 0;
    mcp3301_measurement_t mt = {0, 0.0};
    for(;; t = (((double)(clock() - t_init))/ CLOCKS_PER_SEC)) {
        mt = read_mcp3301_measurement(spi_fd, t_init);
        fb_push(fb, mt.int_val);
        printf("%5.6f\t%d\t%4.3f\n", mt.timestamp, mt.int_val, filter_avg(fb));
        loops++;
    }
    printf("Loops: %d\tAvg time: %2.8f\tLoops/sec: %d\n", loops, t / loops, (int)(loops / t));

    // do stuff

    fb_del(fb);
    spi_shutdown(spi_fd);
    return EXIT_SUCCESS;

fail:
    printf("Done with errors\n");
    return EXIT_FAILURE;
}

